package com.chelvc.framework.security.token;

import java.util.Date;
import java.util.concurrent.TimeUnit;

import com.chelvc.framework.base.context.SessionContextHolder;
import com.chelvc.framework.base.model.Session;
import com.chelvc.framework.base.model.Terminal;
import com.chelvc.framework.security.util.TokenUtils;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.security.oauth2.common.OAuth2AccessToken;
import org.springframework.security.oauth2.provider.OAuth2Authentication;
import org.springframework.security.oauth2.provider.token.store.JwtAccessTokenConverter;
import org.springframework.security.oauth2.provider.token.store.JwtTokenStore;

/**
 * 基于Redis的令牌存储实现
 *
 * @author Woody
 * @date 2023/4/5
 */
@Slf4j
public class RedisTokenStore extends JwtTokenStore {
    private final RedisTemplate<String, Object> redisTemplate;

    public RedisTokenStore(@NonNull RedisTemplate<String, Object> redisTemplate,
                           @NonNull JwtAccessTokenConverter jwtAccessTokenConverter) {
        super(jwtAccessTokenConverter);
        this.redisTemplate = redisTemplate;
    }

    /**
     * 保存令牌信息
     *
     * @param terminal   客户终端
     * @param id         主体标识
     * @param token      令牌标识
     * @param expiration 过期时间
     */
    protected void save(Terminal terminal, Long id, String token, Date expiration) {
        String key = TokenUtils.key(terminal, id);
        if (log.isDebugEnabled()) {
            log.debug("Saving redis token key: {}", key);
        }
        long timeout = expiration.getTime() - System.currentTimeMillis();
        this.redisTemplate.opsForValue().set(key, token, timeout, TimeUnit.MILLISECONDS);
    }

    /**
     * 移除令牌
     *
     * @param terminal 客户终端
     * @param id       主体标识
     */
    protected void remove(Terminal terminal, Long id) {
        String key = TokenUtils.key(terminal, id);
        if (log.isDebugEnabled()) {
            log.debug("Removing redis token key: {}", key);
        }
        this.redisTemplate.delete(key);
    }

    @Override
    public void storeAccessToken(OAuth2AccessToken token, OAuth2Authentication authentication) {
        Terminal terminal = SessionContextHolder.getSessionProperty(Session::getTerminal);
        this.save(terminal, TokenUtils.getId(token), token.getValue(), token.getExpiration());
    }

    @Override
    public void removeAccessToken(OAuth2AccessToken token) {
        Terminal terminal = SessionContextHolder.getSessionProperty(Session::getTerminal);
        this.remove(terminal, TokenUtils.getId(token));
    }
}
